import aiohttp
import pytest
from sharkiq.ayla_api import get_ayla_api, AylaApi
from sharkiq.const import SHARK_APP_ID, SHARK_APP_SECRET, AUTH0_CLIENT_ID
from sharkiq.exc import (
    SharkIqAuthError,
    SharkIqAuthExpiringError,
    SharkIqNotAuthedError,
    AUTH_EXPIRED_MESSAGE,
    NOT_AUTHED_MESSAGE,
)
from datetime import datetime, timedelta


def test_get_ayla_api():
    api = get_ayla_api("myusername@mysite.com", "mypassword")

    assert api._email == "myusername@mysite.com"
    assert api._password == "mypassword"
    assert api._access_token is None
    assert api._refresh_token is None
    assert api._auth_expiration is None
    assert api._is_authed == False
    assert api._app_id == SHARK_APP_ID
    assert api._app_secret == SHARK_APP_SECRET
    assert api.websession is None


class TestAylaApi:
    def test_init__required_vals(self):
        api = AylaApi(
            "myusername@mysite.com", "mypassword", "app_id_123", "client_id_123", "appsecret_123"
        )

        assert api._email == "myusername@mysite.com"
        assert api._password == "mypassword"
        assert api._access_token is None
        assert api._refresh_token is None
        assert api._auth_expiration is None
        assert api._is_authed == False
        assert api._app_id == "app_id_123"
        assert api._app_secret == "appsecret_123"
        assert api._auth0_client_id == "client_id_123"
        assert api.websession is None

    @pytest.mark.asyncio
    async def test_ensure_session(self, dummy_api):
        # Initially created with no websession
        assert dummy_api.websession is None

        session = await dummy_api.ensure_session()

        # Check that session was created and returned
        assert isinstance(session, aiohttp.ClientSession)
        assert dummy_api.websession is session

    def test_property__login_data(self, dummy_api):
        assert dummy_api._login_data["app_id"] == SHARK_APP_ID
        assert dummy_api._login_data["app_secret"] == SHARK_APP_SECRET

    def test_auth0__login_data(self, dummy_api):
        assert dummy_api._auth0_login_data == {
            "grant_type":"password",
            "client_id": AUTH0_CLIENT_ID,
            "username": "myusername@mysite.com",
            "password": "mypassword",
            "scope": "openid profile email offline_access"
        }

    def test_set_id_token__401_requires_verification_response(self, dummy_api):
        with pytest.raises(SharkIqAuthError) as e:
            dummy_api._set_id_token(401, {"error": "requires_verification", "error_description": "description"})
        assert e.value.args[0] == "description. Auth request flagged for verification."

    def test_set_id_token__401_response(self, dummy_api):
        with pytest.raises(SharkIqAuthError) as e:
            dummy_api._set_id_token(401, {"error": "generic", "error_description": "generic"})
        assert e.value.args[0] == "generic. Confirm credentials are correct."

    def test_set_id_token__400_response(self, dummy_api):
        with pytest.raises(SharkIqAuthError) as e:
            dummy_api._set_id_token(400, {"error": "generic", "error_description": "generic"})
        assert e.value.args[0] == "generic"

    def test_set_id_token__403_response(self, dummy_api):
        with pytest.raises(SharkIqAuthError) as e:
            dummy_api._set_id_token(403, {"error": "generic", "error_description": "generic"})
        assert e.value.args[0] == "generic"
        
    def test_set_credentials__404_response(self, dummy_api):
        with pytest.raises(SharkIqAuthError) as e:
            dummy_api._set_credentials(404, {"errors": "Not found"})
        assert (
            e.value.args[0] == "Not found (Confirm app_id and app_secret are correct)"
        )

    def test_set_credentials__401_response(self, dummy_api):
        with pytest.raises(SharkIqAuthError) as e:
            dummy_api._set_credentials(401, {"errors": "Unauthorized"})
        assert e.value.args[0] == "Unauthorized"

    def test_set_credentials__valid_response(self, dummy_api):
        assert dummy_api._access_token is "token123"
        assert dummy_api._refresh_token is "token321"
        assert dummy_api._auth_expiration.timestamp() == pytest.approx(
            (datetime.now() + timedelta(seconds=700)).timestamp()
        )
        assert dummy_api._is_authed == True

        t1 = datetime.now() + timedelta(seconds=3600)
        dummy_api._set_credentials(
            200,
            {
                "access_token": "token123",
                "refresh_token": "token321",
                "expires_in": 3600,
            },
        )

        assert dummy_api._access_token == "token123"
        assert dummy_api._refresh_token == "token321"
        assert dummy_api._auth_expiration.timestamp() == pytest.approx(t1.timestamp())
        assert dummy_api._is_authed == True

    def test_property__sign_out_data(self, dummy_api):
        assert dummy_api.sign_out_data == {
            "user": {"access_token": dummy_api._access_token}
        }

    def test_clear_auth(self, dummy_api):
        assert dummy_api._is_authed == True

        dummy_api._clear_auth()

        assert dummy_api._access_token is None
        assert dummy_api._refresh_token is None
        assert dummy_api._auth_expiration is None
        assert dummy_api._is_authed == False

    def test_property__auth_expiration__not_authed(self, dummy_api):
        dummy_api._is_authed = False
        dummy_api._auth_expiration = None

        assert dummy_api.auth_expiration is None

    def test_property__auth_expiration__no_expiration(self, dummy_api):
        # mock the invalid state
        dummy_api._is_authed = True
        dummy_api._auth_expiration = None

        # Check that the correct exception is raised when accessing property
        with pytest.raises(SharkIqNotAuthedError) as e:
            _ = dummy_api.auth_expiration
        assert e.value.args[0] == "Invalid state.  Please reauthorize."

    def test_property__auth_expiration__not_authed(self, dummy_api):
        dummy_api._is_authed = True
        t = datetime.now() + timedelta(seconds=3600)
        dummy_api._auth_expiration = t

        assert dummy_api.auth_expiration == t

    def test_property__token_expired__false(self, dummy_api):
        dummy_api._is_authed = True
        dummy_api._auth_expiration = datetime.now() + timedelta(seconds=3600)
        assert dummy_api.token_expired == False

    def test_property__token_expired__true(self, dummy_api):
        dummy_api._is_authed = True
        dummy_api._auth_expiration = datetime.now() - timedelta(seconds=3600)
        assert dummy_api.token_expired == True

    def test_property__token_expired__not_authed(self, dummy_api):
        dummy_api._is_authed = False
        dummy_api._auth_expiration = datetime.now() + timedelta(seconds=3600)
        assert dummy_api.token_expired == True

    def test_property__token_expiring_soon__false(self, dummy_api):
        dummy_api._is_authed = True
        # "soon" is considered to be within 600 seconds from the current time
        dummy_api._auth_expiration = datetime.now() + timedelta(seconds=605)
        assert dummy_api.token_expiring_soon == False

    def test_property__token_expiring_soon__true(self, dummy_api):
        dummy_api._is_authed = True
        dummy_api._auth_expiration = datetime.now() + timedelta(seconds=595)
        assert dummy_api.token_expiring_soon == True

    def test_property__token_expiring_soon__not_authed(self, dummy_api):
        dummy_api._is_authed = False
        dummy_api._auth_expiration = datetime.now() + timedelta(seconds=3600)
        assert dummy_api.token_expiring_soon == True

    @pytest.mark.parametrize(
        "access_token,auth_state,auth_timedelta",
        [
            ("token123", True, timedelta(seconds=-100)),  # auth expiry passed
            (None, True, timedelta(seconds=700)),  # invalid token
            ("token123", False, timedelta(seconds=-100)),  # not authed
        ],
    )
    def test_check_auth__not_authed(
        self, dummy_api, access_token, auth_state, auth_timedelta
    ):
        dummy_api._access_token = access_token
        dummy_api._is_authed = auth_state
        dummy_api._auth_expiration = datetime.now() + auth_timedelta

        with pytest.raises(SharkIqNotAuthedError) as e:
            dummy_api.check_auth()

        assert e.value.args[0] == NOT_AUTHED_MESSAGE
        assert dummy_api._is_authed == False

    def test_check_auth__expiring_soon_exception(self, dummy_api):
        dummy_api._auth_expiration = datetime.now() + timedelta(seconds=400)

        with pytest.raises(SharkIqAuthExpiringError) as e:
            dummy_api.check_auth(raise_expiring_soon=True)

        assert e.value.args[0] == AUTH_EXPIRED_MESSAGE

        # No exception raised when set to False
        dummy_api.check_auth(raise_expiring_soon=False)

    def test_check_auth__valid(self, dummy_api):
        assert dummy_api.check_auth() is None

    def test_auth_header(self, dummy_api):
        dummy_api._access_token = "myfaketoken"
        assert dummy_api.auth_header == {
            "Authorization": "auth_token myfaketoken"
        }

    def test_get_headers__no_kwargs(self, dummy_api):
        headers = dummy_api._get_headers({})
        assert headers == dummy_api.auth_header

    def test_get_headers__kwargs_(self, dummy_api):
        headers = dummy_api._get_headers({"headers": {"X-Test": "val"}})

        assert headers == {
            "X-Test": "val",
            "Authorization": f"auth_token {dummy_api._access_token}"
        }


"""
Auth0 API router for authentication to the Shark API
"""

import aiohttp
import urllib
from .const import (
    AUTH0_URL,
    EU_AUTH0_URL,
    AUTH0_CLIENT_ID,
    AUTH0_REDIRECT_URI,
    AUTH0_SCOPES
)

from .exc import SharkIqAuthError

class Auth0Client:

    @staticmethod
    async def do_auth0_login(
        session: aiohttp.ClientSession, europe: bool, username: str, password: str
    ) -> dict:
        """Perform Auth0 login like the SharkClean app and return tokens."""

        AUTH_DOMAIN = EU_AUTH0_URL if europe else AUTH0_URL
        CLIENT_ID = AUTH0_CLIENT_ID
        REDIRECT_URI = (AUTH0_REDIRECT_URI)
        SCOPE = AUTH0_SCOPES

        HEADERS = {
            "User-Agent": (
                "Mozilla/5.0 (Linux; Android 10; K) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/139.0.0.0 Mobile Safari/537.36"
            ),
            "Content-Type": "application/x-www-form-urlencoded",
            "Origin": AUTH_DOMAIN,
            "Referer": AUTH_DOMAIN + "/",
        }

        # -------------------
        # Step 1: /authorize
        # -------------------
        authorize_url = (
            f"{AUTH_DOMAIN}/authorize?"
            + urllib.parse.urlencode(
                {
                    "os": "android",
                    "response_type": "code",
                    "client_id": CLIENT_ID,
                    "redirect_uri": REDIRECT_URI,
                    "scope": SCOPE,
                }
            )
        )

        async with session.get(
            authorize_url, headers=HEADERS, allow_redirects=True
        ) as resp:
            parsed = urllib.parse.urlparse(str(resp.url))
            state = urllib.parse.parse_qs(parsed.query).get("state", [None])[0]

        if not state:
            raise SharkIqAuthError("No state returned from /authorize")

        # -------------------
        # Step 2: /u/login
        # -------------------
        login_url = f"{AUTH_DOMAIN}/u/login?state={state}"
        form_data = {
            "state": state,
            "username": username,
            "password": password,
            "action": "default",
        }
        async with session.post(
            login_url, headers=HEADERS, data=form_data, allow_redirects=False
        ) as resp:
            redirect_url = resp.headers.get("Location")

        code = None
        if redirect_url and redirect_url.startswith("/authorize/resume"):
            resume_url = AUTH_DOMAIN + redirect_url
            async with session.get(
                resume_url, headers=HEADERS, allow_redirects=False
            ) as resp:
                final_url = resp.headers.get("Location")
                if final_url:
                    parsed = urllib.parse.urlparse(final_url)
                    code = urllib.parse.parse_qs(parsed.query).get("code", [None])[0]
        else:
            parsed = urllib.parse.urlparse(redirect_url or "")
            code = urllib.parse.parse_qs(parsed.query).get("code", [None])[0]

        # NEW: handle deep link redirect
        if not code and redirect_url and redirect_url.startswith(REDIRECT_URI):
            parsed = urllib.parse.urlparse(redirect_url)
            code = urllib.parse.parse_qs(parsed.query).get("code", [None])[0]

        if not code:
            raise SharkIqAuthError(f"Auth0 login failed: {redirect_url}")


        # -------------------
        # Step 3: /oauth/token
        # -------------------
        token_url = f"{AUTH_DOMAIN}/oauth/token"
        payload = {
            "grant_type": "authorization_code",
            "client_id": CLIENT_ID,
            "code": code,
            "redirect_uri": REDIRECT_URI,
        }
        async with session.post(
            token_url, headers={"Content-Type": "application/json"}, json=payload
        ) as resp:
            token_data = await resp.json()

        if "access_token" not in token_data:
            raise SharkIqAuthError("Auth0 did not return an access token")

        return token_data




import pytest
import os
from sharkiq.ayla_api import get_ayla_api
from datetime import datetime, timedelta

@pytest.fixture
def dummy_api():
    """AylaApi object with invalid auth creds and attributes populated."""
    username = "myusername@mysite.com"
    password = "mypassword"

    dummy_api = get_ayla_api(username=username, password=password)
    dummy_api._access_token = "token123"
    dummy_api._refresh_token = "token321"
    dummy_api._is_authed = True
    dummy_api._auth_expiration = datetime.now() + timedelta(seconds=700)
    return dummy_api


@pytest.fixture
def sample_api():
    """AylaApi object using user-supplied auth creds via SHARKIQ_USERNAME and 
    SHARKIQ_PASSWORD environement variables."""
    username = os.getenv("SHARKIQ_USERNAME")
    password = os.getenv("SHARKIQ_PASSWORD")

    assert username is not None, "SHARKIQ_USERNAME environment variable unset"
    assert password is not None, "SHARKIQ_PASSWORD environment variable unset"

    return get_ayla_api(username=username, password=password)


@pytest.fixture
def sample_api_logged_in(sample_api):
    """Sample API object with user-supplied creds after performing auth flow."""
    sample_api.sign_in()
    return sample_api



import math
import random
import hashlib
import codecs
import base64
import urllib.parse

from .const import (
    AUTH0_URL,
    AUTH0_CLIENT_ID,
    AUTH0_SCOPES,
    AUTH0_REDIRECT_URI,
    EU_AUTH0_URL
)

class FallbackAuth:
  def GenerateFallbackAuthURL(europe: bool):
    """
    Generate an authorization URL for Auth0 that mimics the Shark app's behavior.

    Args:
        europe: If True, use the EU Auth0 URL and app ID/secret.

    Returns:
        The authorization URL.
    """
    state = FallbackAuth.generateRandomString(43)
    verification = FallbackAuth.generateRandomString(43)
    challenge = FallbackAuth.generateChallengeB64Hash(verification)
    base_url = EU_AUTH0_URL if europe == True else AUTH0_URL

    url = (base_url + "/authorize?os=ios&response_type=code&mobile_shark_app_version=rn1.01"
    + '&client_id=' + FallbackAuth.urlEncode(AUTH0_CLIENT_ID)
    + '&state=' + FallbackAuth.urlEncode(state)
    + '&scope=' + FallbackAuth.urlEncode(AUTH0_SCOPES)
    + '&redirect_uri=' + FallbackAuth.urlEncode(AUTH0_REDIRECT_URI)
    + '&code_challenge=' + FallbackAuth.urlEncode(challenge)
    + '&screen_hint=signin'
    + '&code_challenge_method=S256'
    + '&ui_locales=en')
    
    return url

  def generateRandomString(length):
    """
    Generate a random string of alphanumeric characters.

    Args:
        length: The length of the string to generate.

    Returns:
        A random string of alphanumeric characters of the specified length.
    """
    characters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'
    result = ''
    for _ in range(length):
      randomIndex = math.floor(random.random() * len(characters))
      result += characters[randomIndex]

    return result

  def generateChallengeB64Hash(verification_code):
    """
    Generate a challenge hash for the PKCE flow.

    Args:
        verification_code: The verification code to use in the hash.

    Returns:
        str: The challenge hash, base64 encoded with URL safe characters.
    """
    verification_encoded = codecs.encode(verification_code, 'utf-8')
    verification_sha256 = hashlib.sha256(verification_encoded)
    challenge_b64 = base64.b64encode(verification_sha256.digest()).decode()
    challenge_b64_clean = challenge_b64.replace("+", "-").replace("/", "_").replace("=", "").replace("$", "")

    return challenge_b64_clean

  def urlEncode(s):
    """
    URL encode a string.

    Args:
        s (str): The string to URL encode.

    Returns:
        str: The URL encoded string.
    """
    return urllib.parse.quote_plus(s)


"""Unofficial SDK for Shark IQ robot vacuums, designed primarily to support an integration for Home Assistant."""

from .ayla_api import AylaApi, get_ayla_api, Auth0Client
from .exc import (
    SharkIqError,
    SharkIqAuthExpiringError,
    SharkIqNotAuthedError,
    SharkIqAuthError,
    SharkIqReadOnlyPropertyError,
)
from .sharkiq import OperatingModes, PowerModes,  Properties, SharkIqVacuum

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    # Python < 3.8
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version("sharkiq")
except PackageNotFoundError:
    # Package is not installed
    __version__ = "unknown"



"""
Simple implementation of the Ayla networks API

Shark IQ robots use the Ayla networks IoT API to communicate with the device.  Documentation can be
found at:
    - https://developer.aylanetworks.com/apibrowser/
    - https://docs.aylanetworks.com/cloud-services/api-browser/
"""

import aiohttp
import requests

from auth0.authentication import GetToken
from auth0.asyncify import asyncify
from datetime import datetime, timedelta
from typing import Dict, List, Optional
from .auth0 import Auth0Client
from .const import (
    DEVICE_URL,
    LOGIN_URL,
    AUTH0_HOST,
    SHARK_APP_ID,
    SHARK_APP_SECRET,
    AUTH0_URL,
    AUTH0_CLIENT_ID,
    AUTH0_SCOPES,
    BROWSER_USERAGENT,
    EU_DEVICE_URL,
    EU_AUTH0_HOST,
    EU_LOGIN_URL,
    EU_SHARK_APP_ID,
    EU_SHARK_APP_SECRET,
    EU_AUTH0_URL,
    EU_AUTH0_CLIENT_ID
)
from .exc import SharkIqAuthError, SharkIqAuthExpiringError, SharkIqNotAuthedError
from .sharkiq import SharkIqVacuum
from .fallback_auth import FallbackAuth

_session = None


def get_ayla_api(username: str, password: str, websession: Optional[aiohttp.ClientSession] = None, europe: bool = False):
    """
    Get an AylaApi object.

    Args:
        username: The email address of the user.
        password: The password of the user.
        websession: A websession to use for the API.  If None, a new session will be created.
        europe: If True, use the EU login URL and app ID/secret.

    Returns:
        An AylaApi object.
    """
    if europe:
        return AylaApi(username, password, EU_SHARK_APP_ID, EU_AUTH0_CLIENT_ID, EU_SHARK_APP_SECRET, websession=websession, europe=europe)
    else:
        return AylaApi(username, password, SHARK_APP_ID, AUTH0_CLIENT_ID, SHARK_APP_SECRET, websession=websession)


class AylaApi:
    """Simple Ayla Networks API wrapper."""

    def __init__(
            self,
            email: str,
            password: str,
            app_id: str,
            auth0_client_id: str,
            app_secret: str,
            websession: Optional[aiohttp.ClientSession] = None,
            europe: bool = False):
        """
        Initialize the AylaApi object.

        Args:
            email: The email address of the user.
            password: The password of the user.
            app_id: The app ID of the Ayla app.
            app_secret: The app secret of the Ayla app.
            websession: A websession to use for the API.  If None, a new session will be created.
            europe: If True, use the EU login URL and app ID/secret.
        """
        self._email = email
        self._password = password
        self._auth0_id_token = None  # type: Optional[str]
        self._access_token = None  # type: Optional[str]
        self._refresh_token = None  # type: Optional[str]
        self._auth_expiration = None  # type: Optional[datetime]
        self._is_authed = False  # type: bool
        self._app_id = app_id
        self._auth0_client_id = auth0_client_id
        self._app_secret = app_secret
        self.websession = websession
        self.europe = europe

    async def ensure_session(self) -> aiohttp.ClientSession:
        """
        Ensure that we have an aiohttp ClientSession.
        
        Returns:
            An aiohttp ClientSession.
        """
        if self.websession is None:
            self.websession = aiohttp.ClientSession()
        return self.websession

    @property
    def _login_data(self) -> Dict[str, Dict]:
        """
        Prettily formatted data for the login flow.
        
        Returns:
            A dict containing the login data.
        """
        return {
            "app_id": self._app_id,
            "app_secret": self._app_secret,
            "token": self._auth0_id_token
        }
    
    @property
    def _auth0_login_data(self) -> Dict[str, Dict]:
        """
        Prettily formatted data for the Auth0 login flow.
        
        Returns:
            A dict containing the login data.
        """
        return {
            "grant_type": "password",
            "client_id": self._auth0_client_id,
            "username": self._email,
            "password": self._password,
            "scope": AUTH0_SCOPES
        }
    
    @property
    def _auth0_login_headers(self) -> Dict[str, Dict]:
        """
        Headers for the Auth0 login flow.
        
        Returns:
            A dict containing the headers to send for the Auth0 login flow.
        """
        return {
            "Accept": "application/json, text/plain, */*",
            "Accept-Encoding": "gzip, deflate, br, zstd",
            "Accept-Language": "en-US,en;q=0.9",
            "Content-Type": "application/json",
            "dnt": "1",
            "Host": EU_AUTH0_HOST if self.europe else AUTH0_HOST,
            "Origin": EU_AUTH0_URL if self.europe else AUTH0_URL,
            "Priority": "u=1, i",
            "Referrer": EU_AUTH0_URL if self.europe else AUTH0_URL + "/",
            "Sec-Ch-Ua": '"Chrome";v="137", "Chromium";v="137", "Not A;Brand";v="24"',
            "Sec-Ch-Ua-Mobile": "?0",
            "Sec-Ch-Ua-Platform": '"macOS"',
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-origin",
            "Sec-Gpc": "1",
            "User-Agent": BROWSER_USERAGENT
        }
    
    @property
    def _ayla_login_headers(self) -> Dict[str, Dict]:
        """
        Headers for the Ayla login flow.
        
        Returns:
            A dict containing the headers to send for the Ayla login flow.
        """
        return {
            "Content-Type": "application/json",
            "User-Agent": BROWSER_USERAGENT
        }

    def _set_credentials(self, status_code: int, login_result: Dict):
        """
        Update the internal credentials store.
        
        Args:
            status_code: The status code of the login response.
            login_result: The result of the login response.
        """
        if status_code == 404:
            raise SharkIqAuthError(login_result["errors"] + " (Confirm app_id and app_secret are correct)")
        elif status_code == 401:
            raise SharkIqAuthError(login_result["errors"])

        self._access_token = login_result["access_token"]
        self._refresh_token = login_result["refresh_token"]
        self._auth_expiration = datetime.now() + timedelta(seconds=login_result["expires_in"])
        self._is_authed = (status_code < 400)

    def _set_id_token(self, status_code: int, login_result: Dict):
        """
        Update the ID token.

        Args:
            status_code: The status code of the login response.
            login_result: The result of the login response.
        """
        if status_code == 401 and login_result["error"] == "requires_verification":

            raise SharkIqAuthError(login_result["error_description"] + ". Auth request flagged for verification.")
        elif status_code == 401:
            raise SharkIqAuthError(login_result["error_description"] + ". Confirm credentials are correct.")
        elif status_code == 400 or status_code == 403:
            raise SharkIqAuthError(login_result["error_description"])
        
        self._auth0_id_token = login_result["id_token"]

    async def async_set_cookie(self):
        """
        Query Auth0 to set session cookies [required for Auth0 support]
        """
        initial_url = self.gen_fallback_url()
        ayla_client = await self.ensure_session()

        async with ayla_client.get(initial_url, allow_redirects=False, headers=self._auth0_login_headers) as auth0_resp:
            ayla_client.cookie_jar.update_cookies(auth0_resp.cookies)

    async def async_sign_in(self, use_auth0=False):
        """
        Authenticate to Ayla API asynchronously via Auth0 [requires cookies]
        """
        ayla_client = await self.ensure_session()

        try:
            if use_auth0 or self.europe:
                AsyncGetToken = asyncify(GetToken)
                get_token = AsyncGetToken(EU_AUTH0_HOST if self.europe else AUTH0_HOST, EU_AUTH0_CLIENT_ID if self.europe else AUTH0_CLIENT_ID)

                auth_result = await get_token.login_async(
                    username=self._email,
                    password=self._password,
                    grant_type='password',
                    scope=AUTH0_SCOPES
                )

                self._auth0_id_token = auth_result["id_token"]
            else:
                auth_result = await Auth0Client.do_auth0_login(
                    ayla_client,
                    self.europe,
                    self._email,
                    self._password
                )
                self._auth0_id_token = auth_result["id_token"]
        except Exception:
            if use_auth0 == False:
                # Try fallback Auth0 API if default method failed
                await self.async_sign_in(True)

        login_data = self._login_data
        login_url = f"{EU_LOGIN_URL if self.europe else LOGIN_URL}/api/v1/token_sign_in"
        async with ayla_client.post(login_url, json=login_data, headers=self._ayla_login_headers) as login_resp:
            login_resp_json = await login_resp.json()
            self._set_credentials(login_resp.status, login_resp_json)


    async def async_refresh_auth(self):
        """
        Refresh the authentication synchronously.
        """
        refresh_data = {"user": {"refresh_token": self._refresh_token}}
        ayla_client = await self.ensure_session()
        async with ayla_client.post(f"{EU_LOGIN_URL if self.europe else LOGIN_URL:s}/users/refresh_token.json", json=refresh_data, headers=self._ayla_login_headers) as resp:
            self._set_credentials(resp.status, await resp.json())

    @property
    def sign_out_data(self) -> Dict:
        """
        Payload for the sign_out call.
        
        Returns:
            A dict containing the sign out data.
        """
        return {"user": {"access_token": self._access_token}}

    def _clear_auth(self):
        """
        Clear authentication state.
        """
        self._is_authed = False
        self._access_token = None
        self._refresh_token = None
        self._auth_expiration = None

    async def async_sign_out(self):
        """
        Sign out and invalidate the access token.
        """
        ayla_client = await self.ensure_session()
        async with ayla_client.post(f"{EU_LOGIN_URL if self.europe else LOGIN_URL:s}/users/sign_out.json", json=self.sign_out_data) as _:
            pass
        self._clear_auth()

    def gen_fallback_url(self):
        """
        Generate a URL for the fallback authentication flow.
        
        Returns:
            The URL for the fallback authentication flow.
        """
        return FallbackAuth.GenerateFallbackAuthURL(self.europe)

    @property
    def auth_expiration(self) -> Optional[datetime]:
        """
        Get the time at which the authentication expires.
        
        Returns:
            The time at which the authentication expires.
        """
        if not self._is_authed:
            return None
        elif self._auth_expiration is None:  # This should not happen, but let's be ready if it does...
            raise SharkIqNotAuthedError("Invalid state.  Please reauthorize.")
        else:
            return self._auth_expiration

    @property
    def token_expired(self) -> bool:
        """
        Return true if the token has already expired.
        
        Returns:
            True if the token has already expired.
        """
        if self.auth_expiration is None:
            return True
        return datetime.now() > self.auth_expiration

    @property
    def token_expiring_soon(self) -> bool:
        """
        Return true if the token will expire soon.
        
        Returns:
            True if the token will expire soon.
        """
        if self.auth_expiration is None:
            return True
        return datetime.now() > self.auth_expiration - timedelta(seconds=600)  # Prevent timeout immediately following

    def check_auth(self, raise_expiring_soon=True):
        """
        Confirm authentication status.
        
        Args:
            raise_expiring_soon: Raise an exception if the token will expire soon.

        Raises:
            SharkIqAuthExpiringError: If the token will expire soon.
            SharkIqAuthError: If the token has already expired.
        """
        if not self._access_token or not self._is_authed or self.token_expired:
            self._is_authed = False
            raise SharkIqNotAuthedError()
        elif raise_expiring_soon and self.token_expiring_soon:
            raise SharkIqAuthExpiringError()

    @property
    def auth_header(self) -> Dict[str, str]:
        """
        Get the authorization header.

        Returns:
            The authorization header.
        """
        self.check_auth()
        return {"Authorization": f"auth_token {self._access_token:s}"}

    def _get_headers(self, fn_kwargs) -> Dict[str, str]:
        """
        Extract the headers element from fn_kwargs, removing it if it exists
        and updating with self.auth_header.

        Args:
            fn_kwargs: The kwargs passed to the function.

        Returns:
            The headers.
        """
        try:
            headers = fn_kwargs['headers']
        except KeyError:
            headers = {}
        else:
            del fn_kwargs['headers']
        headers.update(self.auth_header)
        return headers

    def request(self, method: str, url: str, **kwargs) -> requests.Response:
        """
        Make a request to the Ayla API.

        Args:
            method: The HTTP method to use.
            url: The URL to request.
            **kwargs: Additional keyword arguments to pass to requests.

        Returns:
            The response from the request.
        """
        headers = self._get_headers(kwargs)
        return requests.request(method, url, headers=headers, **kwargs)

    async def async_request(self, http_method: str, url: str, **kwargs):
        """
        Make a request to the Ayla API.
        
        Args:
            http_method: The HTTP method to use.
            url: The URL to request.
            **kwargs: Additional keyword arguments to pass to requests.

        Returns:
            The response from the request.
        """
        ayla_client = await self.ensure_session()
        headers = self._get_headers(kwargs)
        result = ayla_client.request(http_method, url, headers=headers, **kwargs)

        return result

    def list_devices(self) -> List[Dict]:
        """
        List the devices on the account.

        Returns:
            A list of devices.
        """
        resp = self.request("get", f"{EU_DEVICE_URL if self.europe else DEVICE_URL:s}/apiv1/devices.json")
        devices = resp.json()
        if resp.status_code == 401:
            raise SharkIqAuthError(devices["error"]["message"])
        return [d["device"] for d in devices]

    async def async_list_devices(self) -> List[Dict]:
        """
        List the devices on the account.

        Returns:
            A list of devices.
        """
        async with await self.async_request("get", f"{EU_DEVICE_URL if self.europe else DEVICE_URL:s}/apiv1/devices.json") as resp:
            devices = await resp.json()
            if resp.status == 401:
                raise SharkIqAuthError(devices["error"]["message"])
        return [d["device"] for d in devices]

    def get_devices(self, update: bool = True) -> List[SharkIqVacuum]:
        """
        Get the devices on the account.
        
        Args:
            update: Update the device list if it is out of date.

        Returns:
            A list of devices.
        """
        devices = [SharkIqVacuum(self, d, europe=self.europe) for d in self.list_devices()]
        if update:
            for device in devices:
                device.get_metadata()
                device.update()
        return devices

    async def async_get_devices(self, update: bool = True) -> List[SharkIqVacuum]:
        """
        Get the devices on the account.

        Args:
            update: Update the device list if it is out of date.
        
        Returns:
            A list of devices.
        """
        devices = [SharkIqVacuum(self, d, europe=self.europe) for d in await self.async_list_devices()]
        if update:
            for device in devices:
                await device.async_get_metadata()
                await device.async_update()
        return devices
    
    async def async_close_session(self):
        """
        Close the shared aiohttp ClientSession.

        This should be called when you are finished with the AylaApi object.
        """
        shared_session = self.ensure_session()
        if shared_session is not None:
            shared_session.close()




"""Various constants"""
AUTH0_URL = "https://login.sharkninja.com"
AUTH0_HOST = "login.sharkninja.com"
AUTH0_CLIENT_ID = "wsguxrqm77mq4LtrTrwg8ZJUxmSrexGi"
AUTH0_SCOPES = "openid profile email offline_access"
AUTH0_REDIRECT_URI = "com.sharkninja.shark://login.sharkninja.com/ios/com.sharkninja.shark/callback"
DEVICE_URL = "https://ads-field-39a9391a.aylanetworks.com"
LOGIN_URL = "https://user-field-39a9391a.aylanetworks.com"
SHARK_APP_ID = "ios_shark_prod-3A-id"
SHARK_APP_SECRET = "ios_shark_prod-74tFWGNg34LQCmR0m45SsThqrqs"
SHARK_APP_USERAGENT = "SharkClean/29562 Darwin/24.3.0"
BROWSER_USERAGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Safari/537.36"
EU_AUTH0_URL = "https://logineu.sharkninja.com"
EU_AUTH0_HOST = "logineu.sharkninja.com"
EU_AUTH0_CLIENT_ID = "rKDx9O18dBrY3eoJMTkRiBZHDvd9Mx1I"
EU_DEVICE_URL = "https://ads-eu.aylanetworks.com"
EU_LOGIN_URL = "https://user-field-eu.aylanetworks.com"
EU_SHARK_APP_ID = "android_shark_prod-lg-id"
EU_SHARK_APP_SECRET = "android_shark_prod-xuf9mlHOo0p3Ty5bboFROSyRBlE"



from sharkiq import get_ayla_api, OperatingModes, Properties, PowerModes

USERNAME = 'me@email.com'
PASSWORD = '$7r0nkP@s$w0rD'

ayla_api = get_ayla_api(USERNAME, PASSWORD)
ayla_api.sign_in()

shark_vacs = ayla_api.get_devices()
shark = shark_vacs[0]

shark.update()
shark.set_operating_mode(OperatingModes.START)
shark.return_to_base()



import asyncio
from sharkiq import get_ayla_api, OperatingModes, SharkIqVacuum

USERNAME = 'me@email.com'
PASSWORD = '$7r0nkP@s$w0rD'

async def main(ayla_api) -> SharkIqVacuum:
    await ayla_api.async_sign_in()
        
    shark_vacs = await ayla_api.async_get_devices()
    shark = shark_vacs[0]
    await shark.async_update()
    await shark.async_find_device()
    await shark.async_set_operating_mode(OperatingModes.START)

    return shark


ayla_api = get_ayla_api(USERNAME, PASSWORD)
shark = asyncio.run(main(ayla_api))
